

/*
	Time Course: Module to analyze 3D stacks based on the 
	assumption that the third dimension is time. Each voxel
	thus represents one value I(x,y,t). This module analyzes
	various properties in t, thus converting I(x,y,t) into A(x,y)
*/



/*		MODULE FOR CRYSTAL IMAGE 0.85 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */




#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#include "includeall.h"






#define CABS(X,Y)	(sqrt((X)*(X)+(Y)*(Y)))
#define SWAP(a,b)	tempr=(a);(a)=(b);(b)=tempr
#define FPS			12


image_type workimg;
char function;
GtkWidget *tc_window;
GtkAdjustment *tc_uthresh_adj;
GtkAdjustment *tc_lthresh_adj;
GtkAdjustment *tc_rthresh_adj;
GtkAdjustment *tc_deltat_spin_adj;


float tc_deltat, tc_uthresh, tc_lthresh, tc_rthresh;



/**************************************************

	Some algorithms used in many analysis functions
	
***************************************************/



/* Replaces data[0..2*nn-1] by its discrete Fourier transform, if isign is input as 1; or replaces
// data[0..2*nn-1] by nn times its inverse discrete Fourier transform, if isign is input as - 1.
// data is a complex array of length nn or, equivalently, a real array of length 2*nn. nn MUST
// be an integer power of 2 (this is not checked for!).
// The sequence is real - imag - real - imag - ....
*/

/* WTH?! How old IS this? Sometimes, going back to long-unused code is like entering deeper
	and deeper into an old castle... You pass the more recent buildings that have been added-on 
	in later times and get to the old, old core of the castle... rough-hewn stone, older
	architectural style... dark and drafty hallways...
	well, I guess what I want to say: TODO -- needs to be replaced by FFTW.
*/



void dft_core (double *data, long nn, int isign)

{
unsigned long n,mmax,m,j,istep,i;
double wtemp,wr,wpr,wpi,wi,theta;	/* Double precision for the trigonometric recurrences. */
double tempr,tempi;

	n=nn << 1;
	j=1;
	for (i=1; i<n; i+=2)			/* This is the bit-reversal section of the routine */
	{ 
		if (j > i) 
		{
			SWAP(data[j-1],data[i-1]);	/* Exchange the two complex numbers. */
			SWAP(data[j  ],data[i  ]);
		}
		m=n >> 1;
		while ((m >= 2) && (j > m))
		{
			j -=m;
			m >>= 1;
		}
		j += m;
	}

	/* Here begins the Danielson-Lanczos section of the routine. */
	
	mmax=2;
	while (n > mmax) /* Outer loop executed log 2 nn times. */
	{
		istep=mmax << 1;
		theta=isign*(6.28318530717959/mmax); /* Initialize the trigonometric recurrence. */
		wtemp=sin(0.5*theta);
		wpr = -2.0*wtemp*wtemp;

		wpi=sin(theta);
		wr=1.0;
		wi=0.0;
		for (m=1;m<mmax;m+=2)				/* Here are the two nested inner loops. */
		{
			for (i=m;i<=n;i+=istep) 
			{
				j=i+mmax;					/* This is the Danielson-Lanczos formula:  */
				tempr=wr*data[j-1]-wi*data[j];
				tempi=wr*data[j]+wi*data[j-1];
				data[j-1]=data[i-1]-tempr;
				data[j]=data[i]-tempi;
				data[i-1] += tempr;
				data[i] += tempi;
			}
			wr=(wtemp=wr)*wpr-wi*wpi+wr;	/* Trigonometric recurrence. */
			wi=wi*wpr+wtemp*wpi+wi;
		}
		mmax=istep;
	}
}


/* Compute autocorrelation function or cross correlation function
   of two time series in z. Return time series of length
   img->zmax (caller must free pointer)

	For each set, each lag j, compute for i datapoints
	
	         	 SUM from i=1 to N-j (( x(i)-xmean)*(y(i+j)-ymean) )/(N-j)
		acf(j) = ---------------------------------------------------------
				 SUM from i=1 to N   (( x(i)-xmean)*(y(i  )-ymean) )/(N)

	Note: Cannot handle RGB.
		
*/	
	


float* compute_acf (image_type* img, int xofs, int yofs)
{
float xmean, ymean, usum, lsum;
float* buf;
long xidx, yidx, zlen;
int cnt, cnt2, lag,z;

	/* Pass 1: Determine average values */
	
	xmean = ymean = 0.0; cnt=0;
	zlen = img->xmax * img->ymax;
	for (z=0; z<img->zmax; z++)
	{
		xidx = xofs + z*zlen;
		yidx = yofs + z*zlen;
		switch (img->imgtype)
		{
			case B_CHAR: 
			{
				unsigned char* p = img->data;
				xmean += p[xidx]; ymean += p[yidx]; cnt++;
				break;
			}
			case B_SHORT: 
			{
				short* p = img->data;
				xmean += p[xidx]; ymean += p[yidx]; cnt++;
				break;
			}
			case B_FLOAT: 
			{
				float* p = img->data;
				xmean += p[xidx]; ymean += p[yidx]; cnt++;
				break;
			}
		}
		
		xmean /= cnt; ymean /= cnt;
	}
	
	/* Pass 2: Compute the actual acf */
	
	buf = (float*)calloc (img->zmax, sizeof (float));
	if (!buf) fatal_memerr();
	
	for (lag=0; lag<img->zmax; lag++)
	{
		usum = lsum = 0.0; cnt=0; cnt2=0;
		
		for (z=0; z<img->zmax; z++)
		{
			xidx = xofs + z*zlen;
			yidx = yofs + z*zlen;

			switch (img->imgtype)
			{
				case B_CHAR: 
				{
					unsigned char* p = img->data;
					lsum += p[xidx]*p[yidx]; cnt++;
					if (z+lag < img->zmax) 
					{
						usum += p[xidx]*p[yidx+lag*zlen]; cnt2++;
					}
					break;
				}
				case B_SHORT: 
				{
					short* p = img->data;
					lsum += p[xidx]*p[yidx]; cnt++;
					if (z+lag < img->zmax) 
					{
						usum += p[xidx]*p[yidx+lag*zlen]; cnt2++;
					}
					break;
				}
				case B_FLOAT: 
				{
					float* p = img->data;
					lsum += p[xidx]*p[yidx]; cnt++;
					if (z+lag < img->zmax) 
					{
						usum += p[xidx]*p[yidx+lag*zlen]; cnt2++;
					}
					break;
				}
			}
			
		}
		if (cnt>0) lsum /= cnt; else lsum=0.0;
		if (cnt2>0) usum /= cnt2; else usum=0.0;
		if (lsum!=0.0) buf[lag] = usum/lsum; else buf[lag]=0.0;
	}

	return buf;
}



/***************************************************

	Compute the 1/f slope through a FFT
	
****************************************************/

void one_over_f (image_type* src, image_type* dest)
{
int n;		/* Dataset length. Power of 2, padded with zeroes */
double* xdata; double* ydata; double* rdata;
float* p;
double nyquist,deltaf,r;
int x,y,z;
fitresult fit;
int bx1,by1,bx2,by2;
float min,max;


	/* determine data set length and frequency increment */
	n=1;
	while (n<src->zmax) n+=n;
	nyquist = 0.5/tc_deltat;
	deltaf = nyquist / n;
	xdata = (double*) calloc (n+10,sizeof (double));
	ydata = (double*) calloc (n+10,sizeof (double));
	rdata = (double*) calloc (2*n+10,sizeof (double));

	p = dest->data;
	roi_bbox (src, &bx1, &by1, &bx2, &by2);	/* Get bounding box of ROI if defined */


	dp (1, "Working... ");
	for (y=by1; y<by2; y++)
	{
		dp (1, "%3.0f\b\b\b", 100.0*(y-by1)/(by2-by1));
		for (x=bx1; x<bx2; x++)
		{
			if (in_roi (x,y))
			{
				/* Copy one time series (into z) from the src image
				   into a separate array, ydata. For the transformation,
				   rdata must follow the format real - imag - real - ...
				   where the imaginary part is left zero.
				   Also, the time series is padded with zeroes */
				
				for (z=0; z<2*n; z++) rdata[z]=0;
				min=999999999; max=-min;
				for (z=0; z<src->zmax; z++)
				{
					xdata[z] = z*deltaf;
					r = rdata[2*z] = readbuf_flt (src,x,y,z);
					if (r>max) max=r;
					if (r<min) min=r;
				}
				
				if ((min>=tc_lthresh) && (max<=tc_uthresh))
				{

					/* Now transform ydata, then prepare ydata with the
						magnitude of the FFT */

					dft_core (rdata,n,1);
					for (z=0; z<n/2; z++)
						ydata[z] = sqrt( SQR(rdata[2*z]) + SQR(rdata[2*z+1]) );

					/* Do a log-log fit of magnitude over frequency.
						Cheap hack: We omit DC component and use only
						frequencies up to 0.5*nyquist */

					fit = loglogfit (xdata,ydata,1,n/4);

					/* Stuff the result into dest, no matter what the
					   correlation coefficient is. The Hurst exponent
					   would be (slope-1)/2.
					   Discard ridiculous values */

					if ((fit.H>-10) && (fit.H < 10))
						p[x + dest->xmax*y] = -fit.H;
					else
						p[x + dest->xmax*y] = 0;
				}
				else
					p[x + dest->xmax*y] = 0;

			}
		}
	}
	dp (1, "DONE.\n");
	make_minmax (dest); dest->uid = next_uid++;
		
	free (xdata);
	free (ydata);
	
}






/***************************************************

	Compute the coefficient of variation,
	slice average,
	normalized range (max-min/avg),
	unnormalized range (max-min), or
	normalized fluctuation range
	
	- several nonfractal measures (sd and max-min)
	normalized to obtain robustness against
	changes in amplitude (not robust against
	offsets, though!)
	
****************************************************/



void compute_cv_nrange (image_type* src, image_type* dest, int type)
{
int bx1,by1,bx2,by2;
int x,y,z,n;
float r, Sx,Sxx,avg,sd,min,max,nr;
float* p;

	p = dest->data;

	roi_bbox (src, &bx1, &by1, &bx2, &by2);	/* Get bounding box of ROI if defined */

	dp (1, "Working... ");
	for (y=by1; y<by2; y++)
	{
		dp (1, "%3.0f\b\b\b", 100.0*(y-by1)/(by2-by1));
		for (x=bx1; x<bx2; x++)
		{
			if (in_roi (x,y))
			{
			
				Sx=0; Sxx=0; n=0; imgminmax (src, &max, &min); 	/* note reversed order of min/max */
				for (z=0; z<src->zmax; z++)
				{
					r= readbuf_flt (src,x,y,z);
					Sx+=r; Sxx+=r*r; n++;
					if (r>max) max=r;
					if (r<min) min=r;

				}
				if ((min>=tc_lthresh) && (max<=tc_uthresh))
				{
					avg = Sx/n;
					sd = SDEV(Sx,Sxx,n);
					if (avg==0) nr=0; else nr = (max-min)/avg;
					if (type==0) 
						p[x + dest->xmax*y] = sd/avg;	/* Coeff of variation */
					else if (type==1)
						p[x + dest->xmax*y] = nr;		/* Normalized range */
					else if (type==2)
						p[x + dest->xmax*y] = avg;		/* Slice average */
					else if (type==3)
						p[x + dest->xmax*y] = (max-min);	/* Slice range */
				}
				else
					p[x + dest->xmax*y] = 0;
			}
		}
	}
	dp (1, "DONE.\n");
	make_minmax (dest); dest->uid = next_uid++;
}





/***************************************************+

	CLUSTER ANALYSIS:

	The first slice is assumed to be a clear
	segmented image that can be subjected
	to cluster size ranking. From there, we
	compute the per-slice and per-cluster
	average intensity and output it in
	tabular form.
	The input image remains unchanged.

*****************************************************/



void compute_cluster_average (image_type* src)
{
int bx1,by1,bx2,by2;
int x,y,z,n,num,c,i;
float r, Sx,Sxx,avg,sd,min,max,nr;
float* p;
image_type clustimg;
cluster_type* list;
gchar sbuf[1024];
gchar *report;
float *sumbuf, *cntbuf;


	if (src->zmax < 2)
	{
		ShowMessage ("Image must have at least two slices");
		return;
	}

	/* The first step is to get the cluster table from Slice 0
		so we can allocate memory */

	if (allocate_image (&clustimg, src->imgtype, src->xmax, src->ymax, 1) ==-1)
		fatal_memerr();
	copybuf_z (src, &clustimg, 0,0);		/* Copy slice 0 for clusterization */

	list = calloc (32790, sizeof(cluster_type));
	num=findclust_2d (&clustimg, 0, 32767, 1, 3, 0, list);	/* Clusterize (default params) by size rank */
	dp (1, "Found %d clusters in slice 0\n", num);

	/* Allocate memory for the report and for the per-slice average computation */

	report = calloc (num*src->zmax*10, sizeof (char));
	sumbuf = calloc (num+50, sizeof (float));
	cntbuf = calloc (num+50, sizeof (float));

	/* Begin the report with centroids, to make it easier to find the clusters */

	sprintf (report, "Cluster averages (%d clusters, %d slices): CSV format\n\n", num, src->zmax-1);
	strcat (report, "CentX");
	for (i=1; i<=num; i++)
	{
		sprintf (sbuf, ", %4d", list[i-1].cog_x);
		strcat (report, sbuf);
	}
	strcat (report, "\n");
	strcat (report, "CentY");
	for (i=1; i<=num; i++)
	{
		sprintf (sbuf, ", %4d", list[i-1].cog_y);
		strcat (report, sbuf);
	}
	strcat (report, "\n");


	roi_bbox (src, &bx1, &by1, &bx2, &by2);	/* Get bounding box of ROI if defined */

	for (z=1; z<src->zmax; z++)			/* Each slice is one line of report */
	{
		for (i=0; i<=num; i++)
		{
			sumbuf[i]=0;				/* Reset the accumulators at the start of a slice */
			cntbuf[i]=0;
		}

		for (y=by1; y<by2; y++)
		{
			for (x=bx1; x<bx2; x++)
			{
				if (in_roi (x,y))
				{
					c = ((short*)clustimg.data)[x + clustimg.xmax*y];		/* Cluster number we are working on */
					if (c)													/* Zero = background -- ignore */
					{
						r= readbuf_flt (src,x,y,z);
						sumbuf[c] += r;
						cntbuf[c] += 1;
					}
				}
			}
		}

		/* End accumulating one slice. Print results. */

		sprintf (sbuf, "%4d", z);
		strcat (report, sbuf);
		for (i=1; i<=num; i++)
		{
			if (cntbuf[i])
				sumbuf[i] /= cntbuf[i];
			else
				sumbuf[i]=0;
			sprintf (sbuf, ", %5.1f", sumbuf[i]);
			strcat (report, sbuf);
		}
		strcat (report, "\n");

	}

	ShowLongMessage (report, "Cluster Averages");
	free (sumbuf); free (cntbuf);
	free (report);


}






/***************************************************+

	Dispersion Analysis to obtain Hurst exponent
	
	Dispersion analysis looks at the effect of averaging
	neighboring points. The SD of each time series with
	increasing bin size is drawn over the bin size,
	and a log-log line is fitted. The slope of this line 
	equals H-1.
	
*****************************************************/



void dispanalysis (image_type* src, image_type* dest, float reject_threshold)
{
double Sx,Sxx,av,min,max, r;
int i,j,cnt,binsz,n;
int bx1,by1,bx2,by2;
int x,y,z;
float* p; double* xdata; double* ydata; double* rdata;
fitresult fit;
int oorcnt, validcnt, threshcnt;


	/* Estimate number of data points needed */
	binsz=1; n=0;
	while (binsz <= src->zmax/4)
	{
		binsz+=binsz; n++;
	}
	dp (1, "Dispersion analysis: %d data points in time course\n",n);
	xdata = (double*) calloc (n+1,sizeof (double));
	ydata = (double*) calloc (n+1,sizeof (double));
	rdata = (double*) calloc (src->zmax,sizeof (double));

	p = dest->data;

	roi_bbox (src, &bx1, &by1, &bx2, &by2);	/* Get bounding box of ROI if defined */
	oorcnt=0; validcnt=0; threshcnt=0;
	
	dp (1, "Working... ");
	for (y=by1; y<by2; y++)
	{
		dp (1, "%3.0f\b\b\b", 100.0*(y-by1)/(by2-by1));
		for (x=bx1; x<bx2; x++)
		{
			if (in_roi (x,y))
			{

				/* Speed up the process by copying one time series
				   into a separate array, rdata. Since we expect
				   to have FBM, differentiate to obtain FBN */
				 
				rdata[0]=readbuf_flt(src,x,y,0);
				min=max=rdata[0];
				for (z=1; z<src->zmax; z++)
				{
					r = readbuf_flt (src,x,y,z);
					rdata[z] = r-rdata[z-1];
					if (r>max) max=r;
					if (r<min) min=r;
				}

				if ((min>=tc_lthresh) && (max<=tc_uthresh))
				{
					/* Perform the dispersion analysis */

					binsz=1; n=0;
					while (binsz <= src->zmax/4)
					{
						Sx=0; Sxx=0; cnt=0;
						for (i=0; i<src->zmax; i+=binsz)
						{
							av=0;
							for (j=0; j<binsz; j++) av+=rdata[i+j];
							av/=binsz;
							Sx+=av; Sxx+=av*av; cnt++;
						}
						xdata[n] = binsz;
						ydata[n] = SDEV(Sx,Sxx,cnt);	/* sdev */
						dp (2, "Dispersion analysis: Bin size %d, num points %d, sd=%f\n",
									binsz,cnt,ydata[n]);
						binsz += binsz; n++;
					}
					fit = loglogfit (xdata, ydata, 1,n-1);
					if (fit.r >= reject_threshold)
					{
						p[x + dest->xmax*y] = 1+fit.H; validcnt++;
					}
					else
					{
						p[x + dest->xmax*y] = 0;	/* OOR if not a good fit */
						oorcnt++;
					}
				}
				else
				{
					p[x + dest->xmax*y] = 0;	/* OOR if not within threshold range */
					threshcnt++;
				}
			}
		}
	}
	dp (1, "DONE.\n");
	make_minmax (dest); dest->uid = next_uid++;

	free (xdata);
	free (ydata);
	free (rdata);
	dp (1, "Dispersion analysis: %d valid points; %d rejected; %d not in threshold range\n",
					validcnt, oorcnt, threshcnt);
}



/**************************************************************

	Detrended Fluctuation Analysis. Another way to obtain
	a scaling exponent. DFA splits the data into windows
	(much like rescaled range analysis) and fits a line
	into the data of each window. This line (trend) is subtracted
	from the data in each window (detrended data).
	Next, for each window the root mean square fluctuation of
	the detrended signal is computed and plotted
	over the window length n. A loglog fit yields the
	scaling exponent.
	DFA WORKS ONLY ON FBM, FBN NEEDS TO BE INTEGRATED
	
*/


/* Compute the round mean square fluctuation (detrended)
   of a given section of data */

double compute_rmsfluct (double* x, double* y, int st, int en)
{
fitresult fit;
double F;		/* Resulting rms fluctuation */
int i;

	fit = lsqfit (x,y,st,en);	/* Get trend */
	
	F=0;
	for (i=st; i<=en; i++)
		F += SQR( y[i]-(x[i]*fit.H+fit.intcpt) );
	F = sqrt (F/(en-st+1));
	return F;

}


/* The following code sucessively divides the dataset into halves,
   sharing the middle point if odd number of points.
   src is a 1Dx2 double array containing the time course,
   x and y are the destination arrays where the data pairs
   of length n and rms fluctuation F go. y has to accomodate two
   data points, the sum and the number of points to compute the average.
*/

void recurse_dfa (double* srcx, double* srcy, double* x, double* y, int npts, int st, int en, int level)
{
int n, mid;
double F;



	n = en-st+1;	/* Number of points to cover */
	if (n<=2) return;	/* Break recursion if st==en or n<=2 (can't fit with 2 points) */
	mid = st+n/2;
	if (n & 0x01)	/* odd number of points */
	{
		recurse_dfa (srcx, srcy, x,y,npts, st, mid, level+1);
		recurse_dfa (srcx, srcy, x,y,npts, mid, en, level+1);
	}
	else			/* even number of pts */
	{
		recurse_dfa (srcx, srcy, x,y,npts, st, mid-1, level+1);
		recurse_dfa (srcx, srcy, x,y,npts, mid, en, level+1);
	}
	
	/* Add it up */
	F = compute_rmsfluct(srcx, srcy, st, en);
	dp (2, "Recurse-DFA: Level %d, range %d .. %d  ",level,st,en);
	dp (2, "  RMS fluctuation %f over %d\n",F,n);
	x[level]=n;
	y[level]+=F;
	y[level+npts]+=1;
}


void dfa_core (image_type* src, image_type* dest, double reject_threshold)
{
int i,j,k,cnt;
fitresult fit;
int npts;
double* xdata;
double* ydata;
double* srcx;
double* srcy;
float* p;
int bx1,by1,bx2,by2;
int x,y,z;
double r,min,max,avg;


	/* Estimate number of data points needed and allocate mem */

	npts = (int)(log(src->zmax)/log(2)+5);

	xdata = (double*) calloc (npts,sizeof (double));
	ydata = (double*) calloc (2*npts,sizeof (double));
	srcx = (double*) calloc (src->zmax,sizeof (double));
	srcy = (double*) calloc (src->zmax,sizeof (double));

	p = dest->data; cnt=0;
	roi_bbox (src, &bx1, &by1, &bx2, &by2);	/* Get bounding box of ROI if defined */
	dp (1, "Working... ");

	for (y=by1; y<by2; y++)
	{
		dp (1, "%3.0f\b\b\b", 100.0*(y-by1)/(by2-by1));
		for (x=bx1; x<bx2; x++)
		{
			if (in_roi (x,y))
			{

				/* Copy one time series (into z) from the src image
				   into a separate array, rdata. Since we expect
				   to have FBM, no further integration is needed
				   (as opposed to e.g. dispersion analysis) */
				 
				min = readbuf_flt (src,x,y,0);
				max = min; avg=0;
				for (z=0; z<src->zmax; z++)
				{
					srcx[z] = z;
					r = readbuf_flt (src,x,y,z);
					srcy[z] = r;
					avg+=r;
					if (r>max) max=r;
					if (r<min) min=r;
				}
				avg/= src->zmax;

				if ((min>=tc_lthresh) && (max<=tc_uthresh))
				{
					/* Perform the detrended fluctuation analysis */
					recurse_dfa (srcx, srcy, xdata, ydata, npts, 0,src->zmax-1, 0);
					/* Postprocess the data: divide the sums by the counts
						to obtain averages, then do a log.log fit.
						Copy the resulting slope of the fit into the 
						dest image */
					j=0;
					for (i=0; i<npts; i++)
					{
						k=ydata[npts+i];
						if (k)
						{
							j=i;
							ydata[i] /= ydata[i+npts];
						}
					}
					fit = loglogfit (xdata, ydata, 0, j);

					if (fit.r >= reject_threshold)
					{
						p[x + dest->xmax*y] = fit.H;
						cnt++;
					}
					else
						p[x + dest->xmax*y] = 0;	/* OOR if not a good fit */
				}
				else
					p[x + dest->xmax*y] = 0;	/* OOR if not within threshold selected */

			}
		}
	}
	dp (1, "DONE. %d pixels of %li accepted.\n",cnt,src->xmax*src->ymax);
	make_minmax (dest); dest->uid = next_uid++;

	free (xdata);
	free (ydata);
	free (srcx);
	free (srcy);
	
}


/* AUTOCORRELATION
	Compute the autocorrelation function of the timecourse.
	This results in a 3D stack of the same dimensions as
	the original image which has to be provided in dest */


void acf_core (image_type* src, image_type* dest, double reject_threshold)
{
int cnt,x,y,z;
int bx1,by1,bx2,by2;
long xofs;
float* p;
float* buf;


	p = dest->data; cnt=0;
	roi_bbox (src, &bx1, &by1, &bx2, &by2);	/* Get bounding box of ROI if defined */
	dp (1, "Working... ");

	for (y=by1; y<by2; y++)
	{
		dp (1, "%3.0f\b\b\b", 100.0*(y-by1)/(by2-by1));
		for (x=bx1; x<bx2; x++)
		{
			if (in_roi (x,y))
			{
				xofs = x + y*src->xmax;
				buf = compute_acf (src, xofs,xofs);
				for (z=0; z<src->zmax; z++)
					p[x + dest->xmax*(y + dest->ymax*z)] = buf[z];
				free (buf);
			}
		}
	}

	make_minmax (dest); dest->uid = next_uid++;

}




/***************************************************

	Compute the correlation coefficient of each
	pixel with its 4 neighbors
	
	Correlation is 
	
	( n Sxy - Sx Sy ) / ( n Sxx - Sx Sx ) ( n Syy - Sy Sy )
	
	under the assumption of n>>1, i.e. n approx n-1
	
****************************************************/


void compute_n4_correl (image_type* src, image_type* dest)
{
int bx1,by1,bx2,by2;
int x,y,z,n,i;
float r,s, Sx,Sxx,Sy[4],Syy[4],Sxy[4], cor, max, min;
float* p;

	p = dest->data;

	roi_bbox (src, &bx1, &by1, &bx2, &by2);	/* Get bounding box of ROI if defined */

	bx1 = MAX (1,bx1);
	bx2 = MIN (src->xmax-2, bx2);
	by1 = MAX (1,by1);
	by2 = MIN (src->ymax-2, by2);

	dp (1, "Working... ");
	for (y=by1; y<=by2; y++)
	{
		dp (1, "%3.0f\b\b\b", 100.0*(y-by1)/(by2-by1));
		for (x=bx1; x<=bx2; x++)
		{
			if (in_roi (x,y))
			{
			
				Sx=0; Sxx=0; n=0; min=99999999; max=-min;
				for (i=0; i<4; i++) { Sy[i]=0; Syy[i]=0; Sxy[i]=0; }
				for (z=0; z<src->zmax; z++)
				{
					r= readbuf_flt (src,x,y,z);
					Sx+=r; Sxx+=r*r; n++;
					
					s = readbuf_flt (src,x-1,y,z);
					Sy[0]+=s; Syy[0]+=s*s; Sxy[0]+=r*s;
					s = readbuf_flt (src,x+1,y,z);
					Sy[1]+=s; Syy[1]+=s*s; Sxy[1]+=r*s;
					s = readbuf_flt (src,x,y-1,z);
					Sy[2]+=s; Syy[2]+=s*s; Sxy[2]+=r*s;
					s = readbuf_flt (src,x,y+1,z);
					Sy[3]+=s; Syy[3]+=s*s; Sxy[3]+=r*s;
					
					if (r>max) max=r;
					if (r<min) min=r;

				}
				if ((min>=tc_lthresh) && (max<=tc_uthresh))
				{
					cor = 0;
					for (i=0; i<4; i++)
						cor += 0.25*(n*Sxy[i]-Sx*Sy[i])/((n*Sxx-Sx*Sx)*(n*Syy[i]*Sy[i]*Sy[i]));

					p[x + dest->xmax*y] = cor;
				}
				else
					p[x + dest->xmax*y] = 0;
			}
		}
	}
	dp (1, "DONE.\n");
	make_minmax (dest); dest->uid = next_uid++;


}





/*****************************************************

	Slice-by-slice histogram:
	Compute a 256-bin histogram per slice and arrange
	it for each slice in a 2D fashion.
	Ideally viewed as 3D surface plot.
	

*****************************************************/

#define TC_NUMBINS	256

void tc_slicehisto (image_type *img)
{
int x,y,z, bin;
int bx1,by1,bx2,by2;
image_type temp;
float buf, min, max;
float* p;

	if (allocate_image (&temp, B_FLOAT, TC_NUMBINS, img->zmax, 1)==-1)
		fatal_memerr();

	p = temp.data;
	roi_bbox (img, &bx1, &by1, &bx2, &by2);	/* Get bounding box of ROI if defined */
	imgminmax (img, &min, &max);
		
	for (z=0; z<img->zmax; z++)		/* slice by slice */
	{

		for (y=by1; y<by2; y++)
		{
			for (x=bx1; x<bx2; x++)
			{
				if (in_roi (x,y))
				{
					buf = (readbuf_flt (img,x,y,z)-min)/(max-min);	/* scale to 0...1 */
					bin = (int) (0.5+buf*255);
					if (bin<0) bin=0; else if (bin>255) bin=255;
					p[bin + temp.xmax*z] += 1.0;
				}
			}
		}
	}
	make_minmax (&temp);
	handback (img, &temp);
}


/*****************************************************

	Debug function: Display one pixel timecourse
	and its analysis (hardwired)

*****************************************************/


void tc_display (image_type* img, int x, int y)
{
double *xdata, *ydata, *srcx, *srcy;
int i,j,k,z,npts;
char s[256];
fitresult fit;



	dp (1, "TC: Analyzing data at %d,%d\n",x,y);
	npts = (int)(log(img->zmax)/log(2)+5);
	dp (1, "TC: with z=%d points, we got %d data points\n",img->zmax,npts-5);

	xdata = (double*) calloc (npts,sizeof (double));
	ydata = (double*) calloc (2*npts,sizeof (double));
	srcx = (double*) calloc (img->zmax,sizeof (double));
	srcy = (double*) calloc (img->zmax,sizeof (double));

	for (z=0; z<img->zmax; z++)
	{
		srcx[z]=z;
		srcy[z]=readbuf_flt (img,x,y,z);
	}
				
	/* Perform the detrended fluctuation analysis */

	recurse_dfa (srcx, srcy, xdata, ydata, npts, 0,img->zmax-1, 0);

	/* Postprocess the data: divide the sums by the counts
		to obtain averages, then do a log.log fit.
		Copy the resulting slope of the fit into the 
		dest image */

	j=0;
	for (i=0; i<npts; i++)
	{
		k=ydata[npts+i];
		if (k)
		{
			j=i;
			ydata[i] /= ydata[i+npts];
		}
	}
	fit = loglogfit (xdata, ydata, 0, j);
	multi_graphme_d (xdata,ydata,npts,2, "log(window)", "log(range)", "Detrended Fluctuation Analysis", PLOT_LOG_X|PLOT_LOG_Y);
	sprintf (s, "Log-log fit: 100H=%7.2f, 100r=%5.1f",100.0*fit.H,100.0*fit.r);
	statusbar_display (s);
	
	
}



/*****************************************************

	G U I  and top-level functions
	
******************************************************/


void compute_avg_main ()		/* Slice average */
{
long idx;
float* p1; float* p2;


	if (allocate_image (&workimg, B_FLOAT, mainimg.xmax, mainimg.ymax,1)==-1)
		fatal_memerr ();
	
	compute_cv_nrange (&mainimg, &workimg, 2);
	
	/* Discard the original image and copy the work buffer into the main image for 
		display and analysis, Leave pointers to parent and drawable intact */

	freebuf (&mainimg);
	if (allocate_imgbuf (&mainimg, B_FLOAT, workimg.xmax, workimg.ymax,1)==-1)
		fatal_memerr ();
	
	/* Note. When copying, CV/NR convert to percent. Avg must not do this. */

	p1=workimg.data; p2=mainimg.data;
	for (idx=0; idx<workimg.xmax*workimg.ymax; idx++) p2[idx]=p1[idx];
	make_minmax (&mainimg);
	
	cur_slice=0;
	gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, 0);
	make_imlib_buffer_from_main_image (0);
	redraw (0);
	
	freebuf (&workimg);
}



void compute_1_over_f_main ()
{
long idx;
float* p1; float* p2;


	if (allocate_image (&workimg, B_FLOAT, mainimg.xmax, mainimg.ymax,1)==-1)
		fatal_memerr ();
	
	one_over_f (&mainimg, &workimg);
	
	/* Discard the original image and copy the work buffer into the main image for display and analysis */

	freebuf (&mainimg);
	if (allocate_imgbuf (&mainimg, B_FLOAT, workimg.xmax, workimg.ymax,1)==-1)
		fatal_memerr ();
	
	p1=workimg.data; p2=mainimg.data;
	for (idx=0; idx<workimg.xmax*workimg.ymax; idx++) p2[idx]=p1[idx];
	make_minmax (&mainimg);
	
	cur_slice=0;
	gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, 0);
	make_imlib_buffer_from_main_image (0);
	redraw (0);

	freebuf (&workimg);
}


void compute_cv_main ()
{
long idx;
float* p1; float* p2;


	if (allocate_image (&workimg, B_FLOAT, mainimg.xmax, mainimg.ymax,1)==-1)
		fatal_memerr ();
	
	compute_cv_nrange (&mainimg, &workimg, 0);
	
	/* Discard the original image and copy the work buffer into the main image for display and analysis */

	freebuf (&mainimg);
	if (allocate_imgbuf (&mainimg, B_FLOAT, workimg.xmax, workimg.ymax,1)==-1)
		fatal_memerr ();
	
	p1=workimg.data; p2=mainimg.data;
	for (idx=0; idx<workimg.xmax*workimg.ymax; idx++) p2[idx]=100.0*p1[idx];
	make_minmax (&mainimg);
	
	cur_slice=0;
	gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, 0);
	make_imlib_buffer_from_main_image (0);
	redraw (0);

	freebuf (&workimg);
}


void dispersion_analysis_main ()
{
long idx;
float* p1; float* p2;


	if (allocate_image (&workimg, B_FLOAT, mainimg.xmax, mainimg.ymax,1)==-1)
		fatal_memerr ();
	
	dispanalysis (&mainimg, &workimg, tc_rthresh);
	
	/* Discard the original image and copy the work buffer into the main image for display and analysis */

	freebuf (&mainimg);
	if (allocate_imgbuf (&mainimg, B_FLOAT, workimg.xmax, workimg.ymax,1)==-1)
		fatal_memerr ();
	
	p1=workimg.data; p2=mainimg.data;
	for (idx=0; idx<workimg.xmax*workimg.ymax; idx++) p2[idx]=p1[idx];
	make_minmax (&mainimg);
	
	cur_slice=0;
	gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, 0);
	make_imlib_buffer_from_main_image (0);
	redraw (0);

	freebuf (&workimg);
}


void compute_nrange_main ()
{
long idx;
float* p1; float* p2;


	if (allocate_image (&workimg, B_FLOAT, mainimg.xmax, mainimg.ymax,1)==-1)
		fatal_memerr ();
	
	compute_cv_nrange (&mainimg, &workimg, 1);
	
	/* Discard the original image and copy the work buffer into the main image for display and analysis */

	freebuf (&mainimg);
	if (allocate_imgbuf (&mainimg, B_FLOAT, workimg.xmax, workimg.ymax,1)==-1)
		fatal_memerr ();
	
	p1=workimg.data; p2=mainimg.data;
	for (idx=0; idx<workimg.xmax*workimg.ymax; idx++) p2[idx]=p1[idx];
	make_minmax (&mainimg);
	
	cur_slice=0;
	gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, 0);
	make_imlib_buffer_from_main_image (0);
	redraw (0);

	freebuf (&workimg);
}



void dfa_main ()
{
long idx;
float* p1; float* p2;


	if (allocate_image (&workimg, B_FLOAT, mainimg.xmax, mainimg.ymax,1)==-1)
		fatal_memerr ();
	
	dfa_core (&mainimg, &workimg, tc_rthresh);
	
	/* Discard the original image and copy the work buffer into the main image for display and analysis */

	freebuf (&mainimg);
	if (allocate_imgbuf (&mainimg, B_FLOAT, workimg.xmax, workimg.ymax,1)==-1)
		fatal_memerr ();
	
	p1=workimg.data; p2=mainimg.data;
	for (idx=0; idx<workimg.xmax*workimg.ymax; idx++) p2[idx]=p1[idx];
	make_minmax (&mainimg);
	
	cur_slice=0;
	gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, 0);
	make_imlib_buffer_from_main_image (0);
	redraw (0);

	freebuf (&workimg);
}


void n4_correl_main ()
{
long idx;
float* p1; float* p2;


	if (allocate_image (&workimg, B_FLOAT, mainimg.xmax, mainimg.ymax,1)==-1)
		fatal_memerr ();
	
	compute_n4_correl (&mainimg, &workimg);
	
	/* Discard the original image and copy the work buffer into the main image for display and analysis */

	freebuf (&mainimg);
	if (allocate_imgbuf (&mainimg, B_FLOAT, workimg.xmax, workimg.ymax,1)==-1)
		fatal_memerr ();
	
	p1=workimg.data; p2=mainimg.data;
	for (idx=0; idx<workimg.xmax*workimg.ymax; idx++) p2[idx]=p1[idx];
	make_minmax (&mainimg);
	
	cur_slice=0;
	gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, 0);
	make_imlib_buffer_from_main_image (0);
	redraw (0);

	freebuf (&workimg);
}



void acf_main ()
{


	if (allocate_image (&workimg, B_FLOAT, mainimg.xmax, mainimg.ymax,mainimg.zmax)==-1)
		fatal_memerr ();
	
	acf_core (&mainimg, &workimg, tc_rthresh);
	
	/* Discard the original image and copy the work buffer into the main image for display and analysis */

	handback (&mainimg, &workimg);
	make_imlib_buffer_from_main_image (0);
	redraw (0);
}


/* This is the function called from the macro interpreter */

void timecourse_macro_driver (image_type* src, image_type* workimg, int meth, float lt, float ut)
{

	if (meth==5)
	{
		if (allocate_image (workimg, B_FLOAT, src->xmax, src->ymax, src->zmax)==-1)
			fatal_memerr ();
	}
	else
	{
		if (allocate_image (workimg, B_FLOAT, src->xmax, src->ymax,1)==-1)
			fatal_memerr ();
	}
	
	tc_lthresh=lt; tc_uthresh = ut; tc_deltat=1;

	if (meth==0)
		one_over_f (src, workimg);				/* One over f */
	else if (meth==1)
		compute_cv_nrange (src, workimg, 0);	/* CV */
	else if (meth==2)
		compute_cv_nrange (src, workimg, 1);	/* NR */
	else if (meth==3)
		dispanalysis (src, workimg, 0.9);
	else if (meth==4)
		dfa_core (src, workimg, 0.9);
	else if (meth==5)
		compute_n4_correl (src, workimg);
	else if (meth==6)
		acf_core (src, workimg, 0.9);
	else if (meth==7)
		compute_cv_nrange (src, workimg, 2);	/* Avg */
	
	/* Don't discard the original image here because it's still in a macro variable */

	make_minmax (workimg);
	
}




void on_tc_okbutton_clicked (GtkButton *button, gpointer user_data)
{

	undo_nexus (&mainimg);

	tc_uthresh = tc_uthresh_adj->value;
	tc_lthresh = tc_lthresh_adj->value;
	tc_rthresh = tc_rthresh_adj->value/100;			/* Percent to absolute */
	tc_deltat  = tc_deltat_spin_adj->value/1000;	/* milliseconds to seconds */

	switch (function)
	{
		case 0: 
			compute_avg_main();
			break;
		case 1: 
			compute_cv_main();
			break;
		case 2: 
			compute_nrange_main();
			break;
		case 3: 
			dispersion_analysis_main();
			break;
		case 4: 
			dfa_main();
			break;
		case 5: 
			compute_1_over_f_main ();
			break;
		case 6: 
			n4_correl_main ();
			break;
		case 7: 
			acf_main ();
			break;
		case 8: 
			tc_slicehisto (&mainimg);
			cur_slice=0;
			gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, 0);
			make_imlib_buffer_from_main_image (0);
			break;
		case 9: 
			compute_cluster_average (&mainimg);
			break;
		default:
		{
		}
	}
	gtk_widget_destroy (tc_window);

}


void on_tc_cancelbutton_clicked (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (tc_window);

}


void tc_set_type_menu (GtkWidget *widget, gpointer user_data)
{

	function = *(int*) user_data;
	dp (3, "Changed analysis type to %d\n", function);
	
}


void create_tc_window (const gchar* title)
{
  GtkWidget *tc_table;
  GtkWidget *tc_label1;
  GtkWidget *tc_label2;
  GtkWidget *tc_label3;
  GtkWidget *tc_label4;
  GtkWidget *tc_label5;
  GtkWidget *tc_hseparator1;
  GtkWidget *tc_okbutton;
  GtkWidget *tc_cancelbutton;
  GtkWidget *tc_uthresh_hscale;
  GtkWidget *tc_lthresh_hscale;
  GtkWidget *tc_rthresh_hscale;
  GtkWidget *tc_deltat_spin;
  GtkWidget *tc_hseparator2;
  GtkWidget *tc_typemenu;
  GtkWidget *tc_typemenu_menu;
  GtkWidget *menuitem;
  GtkTooltips *tt_menu1, *tt_menu2, *tt_menu3, *tt_menu4, *tt_menu5, *tt_menu6, *tt_menu7, *tt_menu8, *tt_menu9, *tt_menu10;
  GtkTooltips *tt_deltat, *tt_rthr, *tt_lthr, *tt_uthr;

  float min,max;
  int y;


	imgminmax (&mainimg, &min, &max);

	tc_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (tc_window), "tc_window", tc_window);
	if (title==NULL)
		gtk_window_set_title (GTK_WINDOW (tc_window), "Time Course Analysis");
	else
		gtk_window_set_title (GTK_WINDOW (tc_window), title);

	tc_table = gtk_table_new (6, 2, FALSE);
	gtk_widget_ref (tc_table);
	gtk_object_set_data_full (GTK_OBJECT (tc_window), "tc_table", tc_table,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (tc_table);
	gtk_container_add (GTK_CONTAINER (tc_window), tc_table);

	y=0;
	tc_label5 = MakeLabel (tc_window,tc_table,"Type of analysis","tc_label5",0,y);
	tc_typemenu = Make_Optionmenu (tc_window, tc_table, "tc_typemenu", 1, y);
	tc_typemenu_menu = gtk_menu_new ();

	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "Slice average", (gpointer) &menuopt0);
	tt_menu1 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu1, menuitem, "Pixel-by-pixel average over all slices", NULL);
	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "Coefficient of Variation", (gpointer) &menuopt1);
	tt_menu2 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu2, menuitem, "Pixel-by-pixel StdDev/average", NULL);
	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "Normalized Range", (gpointer) &menuopt2);
	tt_menu3 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu3, menuitem, "Pixel-by-pixel (maximum-minimum)/average", NULL);
	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "Dispersion Analysis", (gpointer) &menuopt3);
	tt_menu4 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu4, menuitem, "Pixel-wise Hurst exponent by 1D dispersion analysis\n(analysis of SD over incresing window sizes)", NULL);
	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "Detrended Fluctuation Analysis", (gpointer) &menuopt4);
	tt_menu5 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu5, menuitem, "Pixel-wise Hurst exponent by DFA\n(RMS fluctuation over increasing window sizes)", NULL);
	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "1/f Spectral Analysis", (gpointer) &menuopt5);
	tt_menu6 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu6, menuitem, "Hurst exponent by 1D frequency spectrum\n(examining log A(f) over -log f and fitting a line)", NULL);
	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "Correlation with neighbors", (gpointer) &menuopt6);
	tt_menu7 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu7, menuitem, "Computes the correlation of each timecourse I(x,y)(z)\nwith its four neighbor timecourses.\nReturns the average of the four correlation values.", NULL);
	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "Autocorrelation function", (gpointer) &menuopt7);
	tt_menu8 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu8, menuitem, "1D autocorrelation, returns a stack of same\nsize as input stack.", NULL);
	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "Slice-by-slice histogram", (gpointer) &menuopt8);
	tt_menu9 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu9, menuitem, "Each line of the output image is the histogram\n(intensity-encoded) of one slice\n of the input image.", NULL);
	menuitem = Option_menuitem (tc_typemenu_menu, GTK_SIGNAL_FUNC(tc_set_type_menu), "Cluster Mean Intensity", (gpointer) &menuopt9);
	tt_menu10 = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_menu10, menuitem, "Compute average intensities by clusters\n(connected pixel groups). The first slice\nMUST contain the cluster mask.\nTabular output, one line per slice.", NULL);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (tc_typemenu), tc_typemenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (tc_typemenu), 0);
	gtk_container_set_border_width (GTK_CONTAINER (tc_typemenu), 5);

	y++;
	tc_label1 = MakeLabel (tc_window,tc_table,"Upper Threshold","tc_label1",0,y);
	tc_uthresh_adj = GTK_ADJUSTMENT(gtk_adjustment_new (max, min, max, 1, 10, 0));
	tc_uthresh_hscale = MakeHScale (tc_uthresh_adj, tc_window, tc_table, "tc_uthresh_hscale",1,y);
	tt_uthr = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_uthr, tc_uthresh_hscale, "Lower and upper bounds for a COMPLETE timecourse\n[I(x,y) over all z] to be included in the analysis.\nUsed in all functions except histogram and cluster intensity. ", NULL);

	y++;
	tc_label2 = MakeLabel (tc_window,tc_table,"Lower Threshold","tc_label2",0,y);
	tc_lthresh_adj = GTK_ADJUSTMENT(gtk_adjustment_new (min, min, max, 1, 10, 0));
	tc_lthresh_hscale = MakeHScale (tc_lthresh_adj, tc_window, tc_table, "tc_lthresh_hscale",1,y);
	tt_lthr = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_lthr, tc_lthresh_hscale, "Lower and upper bounds for a COMPLETE timecourse\n[I(x,y) over all z] to be included in the analysis.\nUsed in all functions except histogram and cluster intensity. ", NULL);

	y++;
	tc_label3 = MakeLabel (tc_window,tc_table,"Rejection Threshold %","tc_label3",0,y);
	tc_rthresh_adj = GTK_ADJUSTMENT(gtk_adjustment_new (90, 0, 110, 1, 10, 0));
	tc_rthresh_hscale = MakeHScale (tc_rthresh_adj, tc_window, tc_table, "tc_rthresh_hscale",1,y);
	tt_rthr = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_rthr, tc_rthresh_hscale, "For regression fits, R-squared value\nbelow which the fit is rejected.\nOnly used in DFA, dispersion, and ACF.", NULL);

	y++;
	tc_label4 = MakeLabel (tc_window,tc_table,"Delta t (msec)","tc_label4",0,y);
	tc_deltat_spin_adj = GTK_ADJUSTMENT(gtk_adjustment_new (100, 0, 100000, 10, 100, 0));
	tc_deltat_spin = MakeSpinbutton (tc_deltat_spin_adj, tc_window, tc_table, "tc_deltat_spin", 1,y);
	tt_deltat = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_deltat, tc_deltat_spin, "Used to calibrate 1/f spectrum\nand its Nyquist frequency.\nUnused otherwise.", NULL);

	y++;
	tc_hseparator1 = MakeHsep (tc_window, tc_table, "tc_hseparator1", 0, y);
	tc_hseparator2 = MakeHsep (tc_window, tc_table, "tc_hseparator1", 1, y);

	tc_okbutton = Button_in_table (tc_window, tc_table, "OK", "tc_okbutton", 0,6, on_tc_okbutton_clicked);
	tc_cancelbutton = Button_in_table (tc_window, tc_table, "Cancel", "tc_cancelbutton", 1,6, on_tc_cancelbutton_clicked);

	gtk_signal_connect (GTK_OBJECT (tc_window), "destroy",
                    	GTK_SIGNAL_FUNC (on_tc_cancelbutton_clicked), NULL);

	gtk_widget_show (tc_window);
}



void timecourse_main ()
{

	function = 0;
	create_tc_window ("Timecourse Analysis");
			
}



/**************************************************************

	This is the macro interpreter function that hooks up
	to the macro interpreter. A pointer to this function gets
	passed to the macro registry function.


**************************************************************/


/*
	timecourse function
	Syntax: imgvar = timecourse (imgvar1, "method", lthresh, uthresh);
	Compute a timecourse image of imgvar1. Imgvar1 must be a stack,
	and each coordinate (x,y) is interpreted as discrete time series
	where the Z coordinate corresponds to the time. One number is computed
	for each time course at (x,y) and returned in imgvar, so that
	imgvar is a 2D image.
	"method" may be any of "1overf", "cv", "nr", "disp", "dfa".
	lthresh and uthresh are the cutoff thresholds at which the timecourse
	is considered clipped.
	Note: It is necessary that the parameter imgvar1 is not an expr,
	because we would have to duplicate the memory, which may cause
	severe problems with large stacks.
*/

variable_type mac_timecourse (unsigned char* s, int* i)
{
variable_type v,r;
int j,k,meth,vidx;
char sparm[20];
float lt, ut;


	/* The image stack */

	j = *i; k=0; v = nullvar;
	if (!isalpha(s[j])) macro_error (1,j);
	do
	{
		v.name[k++] = s[j++];
	} 
	while ((k<16) && (isalpha2(s[j])));
	
	vidx = findvar (v.name);
	if (vidx<0)		/* not found error */
		macro_error (16,j);
	if (vars[vidx].type != MT_IMG) macro_error (6,j);

	/* The timecourse method parameter */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_STRING);	strncpy (sparm, r.strg, 18); free (r.strg);
	meth=0;
	if (!strncasecmp (sparm, "1overf",6)) meth=0;
	else if (!strncasecmp (sparm, "cv",2)) meth=1;
	else if (!strncasecmp (sparm, "nr",2)) meth=2;
	else if (!strncasecmp (sparm, "disp",4)) meth=3;
	else if (!strncasecmp (sparm, "dfa",3)) meth=4;
	else if (!strncasecmp (sparm, "avg",3)) meth=7;
	else macro_error (15,j);	

	/* The thresholds */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); lt = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); ut = r.numvalue;

	while (s[j]!=')') j++;
	*i = j+1;

	/* Go execute. Prepare v to be returned as image */

	v = nullvar;
	timecourse_macro_driver (&vars[vidx].imgptr, &v.imgptr, meth, lt,ut);
	v.type=MT_IMG;

	return v;
}





/**************************************************************

	This is the infrastructure we use to register with the 
	main program. The pointer "menufunction" points to the 
	main entry (the function to be called when the menu is 
	activated). Initmodule is the one function that the 
	main program calls (the name is fixed!) to initialize
	the module and to register the module with the menu.

***************************************************************/

void (*menufunction)();
variable_type (*macrofunction)(unsigned char* s, int* i);


void module_main()
{

	dp (1, "Timecourse module activated\n");
	if (file_ok) timecourse_main ();
}


void initmodule ()
{
	dp (1, "Timecourse module: Initializer called\n");
	
	/* Register this module with the main menu */
	
	menufunction = &module_main;
	dso_menu_register ("Timecourse analysis", menufunction);
	
	/* Register the module with the macro interpreter */
	
	macrofunction = &mac_timecourse;
	register_dso_macro ("TIMECOURSE",4,MT_IMG,macrofunction);
	
}

